/*
        cl.cpp -- Container Lite v 1.89
        (C) Copyright 1995  John Webster Small
        All rights reserved
*/

#include "cl.h"
#include <fstream.h>
#ifndef assert_h
  #include <assert.h>
#endif
#include <stdlib.h>

char * cl:: vD = 0;

char * CL_strdup(const char * D)
{
    if (D)  {
        size_t len = strlen(D);
        char * newD = new char[len+1];
        if (newD)
            return strcpy(newD,D);
    }
    return 0;
}

ostream& CL_puts(ostream& os, const char * D)
{
    size_t len = (D? strlen(D) : 0);
    if ((os << len << endm))
        if (len)
            if (len > INT_MAX)
            {
                (void) os.write(D,INT_MAX);
                (void) os.write(&D[INT_MAX],
                    (int)(len-INT_MAX));
            }
            else
                (void) os.write(D,(int)len);
    return os;
}

istream& CL_gets(istream& is, char *& D)
{
    if (D)  {
        delete[] D;
        D = 0;
    }
    size_t len;
    if ((is >> len >> nextm))
        if (len)  {
          CL_try_new
            D = new char[len+1];
          CL_catch_xalloc(D)
          if (D)  {
            cl::setVD(D);
              // possible dangler
              if (len > INT_MAX)
              {
                (void) is.read(D,INT_MAX);
                (void) is.read(&D[INT_MAX],
                  ((int)len-INT_MAX));
              }
              else
                (void) is.read(D,(int)len);
            if (is)  {
              cl::resetVD();
              (D)[len] = '\0';
            }
            else  {
              delete[] D;
              D = 0;
            }
            return is;
          }
        }
    return is;
}

int  cl:: rcmp(const void * K, const void * D,
        voidKeyCmP kcmP, cl::RelCmp rc)
{
        switch (rc)  {
        case EQ: return   (*kcmP)(K,D);
        case NE: return  !(*kcmP)(K,D);
        case GT: return (((*kcmP)(K,D) <  0)?  0 : 1);
        case GE: return (((*kcmP)(K,D) <= 0)?  0 : 1);
        case LT: return (((*kcmP)(K,D) <= 0)? -1 : 0);
        case LE: return (((*kcmP)(K,D) <  0)? -1 : 0);
        default: assert(0); break;
        }
        return 0;
}

void cl:: init(unsigned flags, unsigned maxNodes,
        unsigned limit, unsigned delta)
{
        nD = 0; vL = 0;
        _curNode = first = _nodes = 0U;
        _cmP = voidCmP0;
/*
        The following relationships are maintained
        during operation of a cl:
        1 <= delta <= lowLimit <= limit <= maxNodes
                <= CL_MAXNODES
        lowThreshold = lowLimit - delta;
*/
        if (maxNodes > CL_MAXNODES)
                maxNodes = CL_MAXNODES;
        if (limit > maxNodes)
                limit = maxNodes;
        if (delta > limit)
                delta = limit;
        if (!delta)  {
                delta = 1;
                if (limit < delta)
                        limit = delta;
                if (maxNodes < limit)
                        maxNodes = limit;
        }
        linkS = (void **)0;
        _limit = limit;
        _delta = delta;
        _maxNodes = maxNodes;
        lowLimit = limit;
        lowThreshold = lowLimit - delta;
        _flags = (flags | CL_SORTED);
}

void * cl:: cloneD(const void * D)
{
    if (D && _flags & CL_NEW)  {
        CL_try_new
          D = newD(D);
        CL_catch_xalloc(D)
        setND((void *)D); // possible dangler
        return (void *)D;
    }
    return 0;
}

void * cl:: decloneD(void * D)
{
    if ( D )
    {
        resetND();
        _flags |= CL_NEWED;
    }
    else
        setND();  // delete cloneD
    return D;
}

void * cl:: deleteLeak(void * OK, void * D )
{
    assert(_flags & CL_DEL);
    if (!OK && D)
        deleteD(D);
    return OK;
}

void cl:: assign(const cl& b, int deep)
{
        allClr();
        (void) setDelta(1);
        (void) setLimit(1);
        (void) setMaxNodes(1);
        (void) setMaxNodes(b.maxNodes());
        (void) setLimit(b.limit());
        (void) setDelta(b.delta());
        resetFlags(CL_ALL_FLAGS);
        setFlags(b.flags());
        if (deep)  {
            setFlags(CL_NEW | CL_DEL);
            for (unsigned i = 0U; i < b.nodes(); i++)
                if (!insQNew(b.atGet(i)))
                        break;
        }
        else
            for (unsigned i = 0U; i < b.nodes(); i++)
                if (!insQ(b.atGet(i)))
                        break;
        setCmP(b.cmP());
        setFlags(b.flags());  // CL_SORTED?
        (void) setCurNode(b.curNode());
}

void cl:: destruct()
{
        if (_flags & CL_DEL)
                (void) allDel();
        else
                allRmv();
}

int cl:: put(ostream& os)
{
        if (!(os << _maxNodes << endm << _limit << endm
                << _delta << endm << _nodes << endm
                << _curNode << endm << _flags << endm
                << cmP2ID(_cmP) << endm))
                return 0;
        if (!putH(os))
                return 0;
        for (unsigned i = 0U; i < _nodes; i++)
                if (!putD(os,atGet(i)))
                        return 0;
        return 1;
}

int cl:: get(istream& is)
{
        unsigned maxNodes, limit, delta, nodes;
        unsigned curNode, flags, cmPid;
        if (!(is >> maxNodes >> nextm >> limit >> nextm
                >> delta >> nextm >> nodes >> nextm
                >> curNode >> nextm >> flags >> nextm
                >> cmPid >> nextm))
                return 0;
        allClr();
        flags |= CL_DEL;
        //reloaded nodes are dynamic!
        setFlags(flags);
        // In case getD() or insQ() throws exception
        // below; curNode setting is lost then!
        (void) setMaxNodes(maxNodes);
        (void) setLimit(limit);
        (void) setDelta(delta);
        setCmP(ID2cmP(cmPid));
        if (!getH(is))
                return 0;
        void * cD, * ok;
        for (unsigned i = 0U; i < nodes; i++)  {
                CL_try_new
                  cD = getD(is);
                CL_catch_xalloc(cD)
                if (!cD) break;
                setND(cD); // possible dangler
                  ok = insQ(cD);
                resetND();
                if (!ok)  {
                        deleteD(cD);
                        break;
                }
        }
        setFlags(flags);
          // CL_SORTED may be raised!
        (void) setCurNode(curNode);
        return (nodes == _nodes);
}

int cl:: load(const char * filename)
{
    int ok = 0;
    if (filename)  {
        ifstream is(filename,ios::in|ios::binary);
        if (is)  {
            ok = get(is);
            is.close();
        }
    }
    return ok;
}

int cl:: save(const char * filename)
{
    int ok = 0;
    if (flags(CL_SAVE) && filename)  {
        ofstream os(filename,ios::out|ios::binary);
        if (os)  {
            ok = put(os);
            os.close();
        }
    }
    return ok;
}

cl:: cl(void ** argv, unsigned argc,
        unsigned flags)
{
   init(flags,CL_MAXNODES,argc,CL_DELTA);
   if (_maxNodes)
        if (argv)
                if (argc > 0U)
                        while (argc--)
                          (void) push(argv[argc]);
                else
                        for (argc = 0U;
                                insQ(argv[argc]);
                                argc++)
                                /* null stmt */;
}

void ** cl:: vector(void ** argv, unsigned argc) const
{
        void ** V;
        if ((V = argv) == (void **)0)  {
                if (!_nodes)
                        return (void **)0;
                CL_try_new
                  V = new void *[argc=_nodes+1];
                CL_catch_xalloc(V)
                if (!V) return (void **)0;
        }
        else if (argc) if (_nodes > argc)
                return (void **)0;
        for (unsigned i = 0U; i < _nodes; i++)
                V[i] = atGet(i);
        while (i < argc)
                V[i++] = (void *)0;
        return V;
}

int cl:: cmp(const cl& b) const
{
    voidCmP cmP_ = (_cmP? _cmP : cmPD());
    if (!cmP_)  {
      cerr << "Assertion failed: cmP(), file "
        << __FILE__ << ", line " << __LINE__ << endl;
      abort();
    }
    int signum;
    for (unsigned i = 0;
        i < nodes() && i < b.nodes();
        i++)
    {
        signum = (*cmP_)(atGet(i),b.atGet(i));
        if (signum)
           return signum;
    }
    return nodes() - b.nodes();
}

unsigned cl:: setLimit(unsigned newLimit)
{
        void ** newLinkS;
        unsigned headNodes;
        if (newLimit < _nodes)
                newLimit = _nodes;
        else if (newLimit > _maxNodes)
                newLimit = _maxNodes;
        if (newLimit < _delta)
                newLimit = _delta;
        if (!newLimit || newLimit == _limit)
                return 0U;
        if (!linkS)
                return (_limit = newLimit);
        CL_try_new
          newLinkS = new void *[newLimit];
        CL_catch_xalloc(newLinkS)
        if (!newLinkS)
                return 0U;
        if ((headNodes = _limit - first) > _nodes)
                headNodes = _nodes;
        if (headNodes)  // move leading nodes
                memcpy(newLinkS,&linkS[first],
                        sizeof(linkS[0U])*headNodes);
        /* copy wrap around trailing nodes */
        if (headNodes < _nodes)
                memcpy(&newLinkS[headNodes],linkS,
                        sizeof(linkS[0U])*
                        (_nodes-headNodes));
        if (newLimit > _limit)
                if (newLimit - _delta > _limit)
                        lowLimit = newLimit - _delta;
                else
                        lowLimit = _limit;
        else
                if (newLimit - _delta > _delta)
                        lowLimit = newLimit - _delta;
                else
                        lowLimit = _delta;
        lowThreshold = lowLimit - _delta;
        delete[] linkS;
        linkS = newLinkS;
        _limit = newLimit;
        first = 0U;
        return _limit;
}

unsigned cl:: setDelta(unsigned newDelta)
{
        return ((newDelta && newDelta <= lowLimit)?
                (_delta = newDelta) : 0U);
}

unsigned cl:: setMaxNodes(unsigned newMaxNodes)
{
        return ((newMaxNodes >= _limit)?
                (_maxNodes = (newMaxNodes
                > CL_MAXNODES)? CL_MAXNODES
                : newMaxNodes) : 0U);
}

void * cl:: atIns(unsigned n, void * D)
{
    void ** newLinkS;
    unsigned newLimit, i;
    int ok;
    if (!D)
        return (void *)0;
    if (_nodes == _limit)  { // no vacancy - expand
        if (_limit == _maxNodes)
            return (void *)0;
        newLimit = (_maxNodes - _delta > _limit)?
            _limit + _delta : _maxNodes;
        CL_try_new
          newLinkS = new void *[newLimit];
        CL_catch_xalloc(newLinkS)
        if (!newLinkS)  // expansion unavailable
                return (void *)0;
        setVL(newLinkS); // possible dangler
          ok = attachD(D);
        resetVL();
        if (!ok)  {// data refuses binding
            delete[] newLinkS;
            return (void *)0;
        }
        newLinkS[((n <= _nodes)? n : (n = _nodes))] = D;
        // calc headNodes
        if (n <= (i = _limit - first))  {
            // insert in head section
            if (n)  {// postfix or split head
                memcpy(newLinkS,&linkS[first],
                    sizeof(linkS[0U])*n);
                if (n < i) // it's a split
                    memcpy(&newLinkS[n+1],
                        &linkS[first+n],
                        sizeof(linkS[0U])*(i-n));
            }
            else    // prefix head
                memcpy(&newLinkS[1],&linkS[first],
                    sizeof(linkS[0U])*i);
            if (i < _nodes)  // copy tail
                memcpy(&newLinkS[1+i],linkS,
                    sizeof(linkS[0U])
                    *(_nodes-i));
        }
        else  {    // insert in tail section
            // copy head first
            memcpy(newLinkS,&linkS[first],
                sizeof(linkS[0U])*i);
            if (i < _nodes)  { // copy tail
                memcpy(&newLinkS[i],linkS,
                    sizeof(linkS[0U])*(n-i));
                if (n < _nodes) // it's a tail split
                    memcpy(&newLinkS[n+1],
                        &linkS[n-i],
                        sizeof(linkS[0U])
                        *(_nodes-n));
            }
        }
        // Compute next smaller linkS size
        // and threshold for shrinking.
        lowLimit = _limit;
        lowThreshold = lowLimit - _delta;
        // swap new for old
        delete[] linkS;
        linkS = newLinkS;
        _limit = newLimit;
        first = 0U;
    }
    else  {  // vacancy
        if (!linkS)
                CL_try_new
                  linkS = new void *[_limit];
                CL_catch_xalloc(linkS)
        if (!linkS)
                return (void *)0;
        if (!attachD(D))  {
            if (!_nodes)  {
                delete[] linkS;
                linkS = (void **)0;
            }
            return (void *)0;
        }
        if (!n)  // push
            linkS[first? --first
                : (first = _limit - 1)] = D;
        else if (n >= _nodes) // insQ
            linkS[(first+(n=_nodes))%_limit] = D;
        else  { // insert interior
            i = (first + n) % _limit;
            if (i < first || !first)
                // move rear rightward
                memmove(&linkS[i+1],&linkS[i],
                    sizeof(linkS[0U])
                    * (_nodes-n));
            else  { // move front leftward
                memmove(&linkS[first-1],
                    &linkS[first],
                    sizeof(linkS[0U])
                    *(n));
                first--;
                i--;
            }
            linkS[i] = D;
        }
    }
    _nodes++;
    if (n <= _curNode)
        _curNode++;
    _flags &= ~CL_SORTED;
    return D;
}

void * cl:: atRmv(unsigned n)
{
    void ** newLinkS = (void **)0;
    unsigned newLimit, i;
    void * D;
    if (!linkS || n >= _nodes)
        return (void *)0;
    D = linkS[i=((first+n)%_limit)];
    if (_nodes <= lowThreshold) {
        // get ready to contract
        newLimit = lowLimit;
            CL_try_new
          newLinkS = new void *[newLimit];
        CL_catch_xalloc(newLinkS)
        // Compute next smaller linkS
        // size and threshold for
        // shrinking.
        if (newLinkS)  {
            if (_nodes < _delta)
                lowThreshold = 0;
            else
                lowThreshold -= _delta;
            lowLimit = lowThreshold + _delta;
        }
    }
    if (newLinkS)  {  // remove and contract
        // calc head nodes
        if ((i = _limit - first) > _nodes)
                i = _nodes;
        if (n < i)  {
            // rmv from head section
            if (n)  { // rmv from head
                memcpy(newLinkS,&linkS[first],
                    sizeof(linkS[0U])*n);
                if (n < i - 1)
                        memcpy(&newLinkS[n],
                            &linkS[first+n+1],
                            sizeof(linkS[0U])*(i-n-1));
            }
            else if (i > 1)   // pop head
                memcpy(newLinkS,&linkS[first+1],
                    sizeof(linkS[0U])*(i-1));
            if (i < _nodes)  // copy tail
                memcpy(&newLinkS[i-1],linkS,
                    sizeof(linkS[0U])
                    *(_nodes-i));
        }
        else  { // i <= n < nodes
            // rmv from tail section
            // copy head first
            memcpy(newLinkS,&linkS[first],
                sizeof(linkS[0U])*i);
            if (i == n)  { // pop tail
                if (n < _nodes - 1)
                    // copy trailing tail
                    memcpy(&newLinkS[i],&linkS[1],
                        sizeof(linkS[0U])*(_nodes-n-1));
            }
            else  {  // copy up to n in tail
                memcpy(&newLinkS[i],linkS,
                    sizeof(linkS[0U])*(n-i));
                // skip n'th node
                if (n < _nodes - 1)
                    // copy anything trailing n'th node
                    memcpy(&newLinkS[n],&linkS[n-i+1],
                        sizeof(linkS[0U])*(_nodes-n-1));
            }
        }
        // swap new for old
        delete[] linkS;
        linkS = newLinkS;
        _limit = newLimit;    //lint !e644
        first = 0U;
    }
    else  { // simply remove
        if (!n)  {  // pop
            if (++first >= _limit)
                first = 0U;
        }
        else if (n != _nodes-1)  // del interior
            if (i < first) // move tail leftward
                memmove(&linkS[i],&linkS[i+1],
                    sizeof(linkS[0U])
                    *(_nodes-n-1));
            else {  // move head rightward
                memmove(&linkS[first+1],
                    &linkS[first],
                    sizeof(linkS[0U])*n);
                first++;
            }
    }
    if (n < _curNode)
        _curNode--;
    else if (n == _curNode)
        _curNode = _nodes-1;
    if (--_nodes == 0U)  {
        _flags |= CL_SORTED;
        _flags &= ~CL_NEWED;
        delete[] linkS;
        linkS = (void **)0;
    }
    detachD(D);
    return D;
}

void cl:: allRmv()
{
//        while (atRmv(0U)) /* null stmt */ ;
    catch_reset();  // delete danglers
    if (linkS)  {
        unsigned i = first, j = _nodes;
        for (i = first, j = _nodes;
            i < _limit && j; i++, j--)
            detachD(linkS[i]);
        for (i = 0; j; i++, j--)
            detachD(linkS[i]);
        delete[] linkS;
    }
    init(_flags,_maxNodes,_delta,_delta);
}

int  cl:: atDel(unsigned n)
{
        void * D;
        if (_flags & CL_DEL)
                if ((D = atRmv(n)) != (void *)0)  {
                        deleteD(D);
                        return 1;
                }
        return 0;
}

int cl:: allDel()
{
    catch_reset();  // delete danglers
    if (_flags & CL_DEL)  {
        if (linkS)  {
            unsigned i = first, j = _nodes;
            for (i = first, j = _nodes;
                i < _limit && j; i++, j--)  {
                detachD(linkS[i]);
                deleteD(linkS[i]);
            }
            for (i = 0; j; i++, j--)  {
                detachD(linkS[i]);
                deleteD(linkS[i]);
            }
            delete[] linkS;
        }
        init(_flags,_maxNodes,_delta,_delta);
        return 1;
    }
    return 0;
}

void * cl:: atPut(unsigned n, void * D)
{
        if (linkS && D && (n < _nodes))  {
                void *& N = linkS[(first+n)%_limit];
                if (D != N) if (attachD(D))  {
                   _flags &= ~CL_SORTED;
                   detachD(N);
                   if (_flags & CL_DEL)
                        deleteD(N);
                   return (N=D);
                }
        }
        return (void *)0;
}

void * cl:: atPutNew(unsigned n, const void * D)
{
        void * oldD, * cD;
        if (!D || !(_flags & CL_NEW))
                return (void *)0;
        if ((oldD = atGet(n)) == (void *)0)
                return (void *)0;
        if (oldD == D)
                return (void *)0;
        CL_try_new
          cD = newD(D);
        CL_catch_xalloc(cD)
        if (cD)  {
                void * ok;
                setND(cD); // possible dangler
                  ok = atPut(n,cD);
                resetND();
                if (ok)  {
                        _flags |= CL_NEWED;
                        return cD;
                }
                else
                        deleteD(cD);
        }
        return (void *)0;
}

void * cl:: atPutAsg(unsigned n, const void * D)
{
        void * oldD;
        if (D && _flags & CL_ASG)
                if ((oldD = atGet(n)) != (void *)0)
                        if (D == oldD)
                                return oldD;
                        else  {
                                _flags &= ~(CL_READ
                                  | CL_READ_DEL);
                                if (assignD(oldD,D))
                                  return oldD;
                        }
        return (void *)0;
}

void * cl:: atGet(unsigned n) const
{
        return ((linkS && (n < _nodes))?
                linkS[(first+n)%_limit] : (void *)0);
}

void * cl:: atGetAsg(unsigned n, void * D)
{
        void * N;
        if (D && _flags & CL_ASG)
                if ((N = atGet(n)) != (void *)0)
                        if (D == N)
                                return D;
                        else  {
                          _flags &= ~CL_READ_DEL;
                          _flags |=  CL_READ;
                          if (assignD(D,N)) return D;
                        }
        return (void *)0;
}

void * cl:: atXchg(unsigned n, void * D)
{
        if (linkS && D && (n < _nodes))
                if (attachD(D))  {
                        _flags &= ~CL_SORTED;
                        void *& N = linkS[(first+n)
                                %_limit];
                        void * oldD = N;
                        N = D;
                        detachD(oldD);
                        return oldD;
                }
        return (void *)0;
}

unsigned cl:: indexOf(const void * D) const
{
        unsigned i;
        if (linkS && D)
                for (i = 0U; i < _nodes; i++)
                        if (D == linkS[(first+i)
                                %_limit])
                                return i;
        return CL_NOTFOUND;
}

void *  cl:: setCurNode(unsigned n)
{
    return ((((_curNode = ((n > _nodes)? _nodes : n))
        < _nodes) && linkS)?
        linkS[(first+_curNode)%_limit]
        : (void *)0);
}

void * cl:: insBefore(void * D)
{ 
    unsigned n = _curNode;
    if (n > _nodes)
        n = _nodes;
    D =  atIns(n,D);
    _curNode = n;
    return D;
}

void * cl:: rmvStayPut()
{ 
    unsigned n = _curNode;
    void * D =  atRmv(n);
    _curNode = n;
    return D;
}

int cl:: delStayPut()
{
    unsigned n = _curNode;
    int ok = atDel(n);
    _curNode = n;
    return ok;
}

int cl:: clrStayPut()
{
    if (_flags & CL_DEL)
        return delStayPut();
    return (rmvStayPut() != 0);
}

void * cl:: insAfter(void * D)
{
        if (atIns(_curNode+1,D))  {
                if (++_curNode >= _nodes)
                        _curNode = _nodes - 1;
                return D;
        }
        return (void *)0;
}

void * cl:: rmvBackup()
{
        void * D;
        unsigned n;
        if ((D = atRmv(n=_curNode)) != (void *)0)
                if (n)
                        _curNode = n - 1;
                else
                        _curNode = _nodes;
        return D;
}

int  cl:: delBackup()
{
        void * D;
        if (_flags & CL_DEL)
                if ((D = rmvBackup()) != (void *)0)  {
                        deleteD(D);
                        return 1;
                }
        return 0;
}

int cl:: clrBackup()
{
    if (_flags & CL_DEL)
        return delBackup();
    return (rmvBackup() != 0);
}

void * cl:: next()
{
        if (linkS)  {
                if (_curNode >= _nodes)
                        _curNode = 0U;
                else
                        _curNode++;
                if (_curNode < _nodes)
                        return linkS[(first+_curNode)
                                %_limit];
        }
        return (void *)0;
}

void * cl:: nextAsg(void * D)
{
    if (_curNode >= _nodes)
        _curNode = 0;  // wrap
    else
        _curNode++;
    if (atGetAsg(_curNode,D))
        return D;
    // if _curNode is a valid node, backup
    if (!_curNode)
        _curNode = _nodes;
    else if (_curNode < _nodes)
        _curNode--;
    return 0;
}

void * cl:: prev()
{
        if (linkS)  {
                if (_curNode)  {
                        if (_curNode > _nodes)
                                _curNode = _nodes;
                        _curNode--;
                }
                else
                        _curNode = _nodes;
                if (_curNode < _nodes)
                        return linkS[(first+_curNode)
                                %_limit];
        }
        return (void *)0;
}

void *  cl:: prevAsg(void * D)
{
    if (_curNode)
    {
        _curNode--;
        if (atGetAsg(_curNode,D))
            return D;
        _curNode++;
    }
    else
        _curNode = _nodes;
    return 0;
}

int   cl:: sort(voidCmP cmP_)
{
        unsigned low, mid, high;
        unsigned d;
        void * D;
/*
        The current node is always reset to undefined
        regardless of the outcome of sort.
*/
        _curNode = _nodes;
        if (!_cmP)
                _cmP = cmPD();
        if (_flags & CL_SORTED)  {
                if (_cmP == cmP_ || !cmP_)
                        return 1;
        }
        else if (!_cmP && !cmP_)
                return 0;
        if (cmP_)  {
                _cmP = cmP_;
                _flags &= ~CL_SORTED;
        }
        if (!_nodes || !linkS)
                return (int) (_flags |= CL_SORTED);
        if (first)  {
                /* form contiguous block at front */
                d = (first + _nodes) % _limit;
                if (d > first)
                        memmove(linkS,&linkS[first],
                            sizeof(linkS[0U])*_nodes);
                else if (d < first)
                        memmove(&linkS[d],
                                &linkS[first],
                                sizeof(linkS[0U])
                                *(_limit-first));
                /* else array is full/contiguous */
                first = 0U;
        }
        for (high = d = 1; d < _nodes; high = ++d)  {
                low = 0U;
                D = linkS[d];
                while (low < high)  {
                        mid = ((high + low) >> 1);
                        if ((*_cmP)(D,
                                linkS[mid]) <= 0)
                                high = mid;
                        else
                                low = mid + 1;
                }
                if (high < d)  {
                        memmove(&linkS[high+1],
                                &linkS[high],
                                sizeof(linkS[0U])
                                *(d-high));
                        linkS[high] = D;
                }
        }
        return (int) (_flags |= CL_SORTED);
}

void * cl:: insSort(void * D)
{
        unsigned low, high;
        void * ok;
/*
        The current node is left undefined if
        anything fails, otherwise it is set to the
        newly inserted node.  If the compare function
        throws an exception, the current node is set
        to the node choked on.
*/
        _curNode = _nodes;
        if (!_cmP) if ((_cmP = cmPD())
                == voidCmP0) return (void *)0;
        if (!D || _nodes >= _maxNodes)
                return (void *)0;
        if (!(_flags & CL_SORTED))
                if (!sort())
                        return (void *)0;
        low = 0U;
        high = _nodes;  // linkS == 0 -> nodes == 0!
        while (low < high)  {
                _curNode = ((high + low) >> 1);
                if ((*_cmP)(D,
                  linkS[(first+_curNode)%_limit]) <= 0)
                  high = _curNode;
                else
                  low = _curNode + 1;
        }
        if ((ok = atIns(high,D)) == (void *)0)
                _curNode = _nodes;
        else
                _curNode = high;
        /*  atIns() resets sorted to zero  */
        _flags |= CL_SORTED;
        return ok;
}

void * cl:: insUnique(void * D)
{
        if (!findFirst(D))
            if (sorted())
                return insSort(D);
            else
                return insQ(D);
        return (void *)0;
}

void * cl:: insUniqueNew(const void * D)
{
        if (!findFirst(D))
            if (sorted())
                return insSortNew(D);
            else
                return insQNew(D);
        return (void *)0;
}

void * cl:: findFirst(const void * K, unsigned& idx,
    voidKeyCmP kcmP, int honorSortFlag, cl::RelCmp rc)
    const
{
        unsigned low, high;
/*
        idx is left undefined if anything fails,
        otherwise it is set to the newly found node.
        If the compare function throws an exception,
        the current node is set to the node choked on.
*/
        idx = _nodes;
        if (!kcmP)  {
          honorSortFlag = 1;
          if ((kcmP = voidKeyCmP(_cmP)) == voidKeyCmP0)
            if ((kcmP = voidKeyCmP(cmPD()))
                == voidKeyCmP0) return (void *) 0;
        }
        if (!linkS || !K || !_nodes)
                return (void *)0;
        if (honorSortFlag && (_flags & CL_SORTED))
          switch (rc)  {
            case EQ:
            case GT:
            case GE:
                /*  binary search for first match  */
                low = 0U;
                high = _nodes - 1;
                while (low < high)  {
                    idx = ((high + low) >> 1);
                    if (rcmp(K,linkS[(first+idx)
                        %_limit],kcmP,rc) <= 0)
                        high = idx;
                    else
                        low = idx + 1;
                }
                idx = high;
                if (rcmp(K,linkS[(first+high)%_limit],
                    kcmP,rc))
                    idx = _nodes;
                break;
            case LT:
            case LE:
                idx = 0;
                if (rcmp(K,linkS[first],kcmP,rc))
                        idx = _nodes;
                break;
            case NE:
                idx = 0;
                if ((*kcmP)(K,linkS[first]))
                        break;
                else
                        idx = _nodes;
                /*  binary search for last match  */
                low = 1U;
                high = _nodes - 1;
                while (low < high)  {
                    idx = ((high + low + 1) >> 1);
                    if ((*kcmP)(K,linkS
                        [(first+idx)%_limit])
                        >= 0)
                        low = idx;
                    else
                        high = idx - 1;
                }
                idx = high;
                if (!(*kcmP)
                        (K,linkS[(first+high)%_limit]))
                        idx = high + 1;
                break;
          }
        else  {  /*  linear search!  */
                idx = 0;
                do  {
                        if (!rcmp(K,linkS
                                [(first+idx)
                                %_limit],kcmP,rc))
                                break;
                } while (++idx < _nodes);
        }
        return ((idx < _nodes)?
                        linkS[(first+idx)%_limit]
                : (void *)0);
}

void * cl:: findNext(const void * K, unsigned& idx,
    voidKeyCmP kcmP, int honorSortFlag, cl::RelCmp rc)
    const
{
/*
        For sorted cl's you must first call
        findFirst() to insure consistent results!
*/
        unsigned low, high;
        void * D;
/*
        idx is left undefined if anything fails,
        otherwise it is set to the newly found node.
        If the compare function throws an exception,
        the current node is set to the node choked on.
*/
        if (!kcmP)  {
          honorSortFlag = 1;
          if ((kcmP = voidKeyCmP(_cmP)) == voidKeyCmP0)
            kcmP = voidKeyCmP(cmPD());
        }
        if (!linkS || !K || !_cmP  || !_nodes
                || (++idx == _nodes))  {
                idx = _nodes;
                return (void *)0;
        }
        if (idx > _nodes)
                idx = 0;
        if (!rcmp(K,D=linkS[(first+idx)%_limit],
                kcmP,rc))
                return D;
        if (honorSortFlag && (_flags & CL_SORTED))  {
                if (rc != NE)
                        return (void *) 0;
                /*  binary search for last match  */
                low  = idx + 1;
                high = _nodes - 1;
                while (low < high)  {
                    idx = ((high + low + 1) >> 1);
                    if ((*kcmP)(K,linkS[(first+idx)
                        %_limit]) >= 0)
                        low = idx;
                    else
                        high = idx - 1;
                }
                idx = high;
                if (!(*kcmP)(K,
                    linkS[(first+high)%_limit]))
                    idx = high + 1;
        }
        else while (++idx < _nodes)
                if (!rcmp(K,linkS[(first+idx)
                        %_limit],kcmP,rc))
                        break;
        return ((idx < _nodes)?
                        linkS[(first+idx)%_limit]
                : (void *)0);
}

//lint -esym(1712,GenericFnCRecord)
struct GenericFnCRecord  {
        GenericFnC _fnC;
        unsigned _id;
        GenericFnCRecord(GenericFnC fnC, unsigned id)
                { _fnC = fnC; _id = id; }
        ~GenericFnCRecord()  {}
};

typedef struct GenericFnCRecord * GenericFnCR;
#define GenericFnCR0  ((GenericFnCR)0)

int FunctionRegistry::regFnC(GenericFnC fnC,
        unsigned id)
{
        unsigned i;
        if (!fnC && !id)  // NULL's id is 0
                return 1;
        if ((!fnC && id) || (fnC && !id))
                return 0;
        for (i = 0U; i < nodes(); i++)
                if (((GenericFnCR)atGet(i))->_id
                        == id)
                        break;
        if (i < nodes())
                        return 0;
        GenericFnCR R;
        CL_try_new
          R = new GenericFnCRecord(fnC,id);
        CL_catch_xalloc(R)
        if (!R)
                return 0;
        else if (!insQ((void *)R))  {
                delete R;
                return 0;
        }
        return 1;
}

unsigned FunctionRegistry::fnC_2_ID(GenericFnC fnC)
{
        GenericFnCR R;
        unsigned i;
        if (!fnC)
                return 0U;
        for (i = 0U; (R = (GenericFnCR) atGet(i))
                != GenericFnCR0; i++)
                if (R->_fnC == fnC)
                        return R->_id;
        return 0U;
}

GenericFnC FunctionRegistry::ID_2_fnC(unsigned id)
{
        GenericFnCR R;
        unsigned i;
        if (!id)
                return GenericFnC0;
        for (i = 0U; (R = (GenericFnCR) atGet(i))
                != GenericFnCR0; i++)
                if (R->_id == id)
                        return R->_fnC;
        return GenericFnC0;
}

FunctionRegistry fnCv;
