/*
        cl.h -- Container Lite v 1.89
        (C) Copyright 1995  John Webster Small
        All rights reserved
*/

#ifndef cl_h
#define cl_h

/* LINTLIBRARY */

///////////////////////////////////////////////////////
//                                                   //
//   If your compiler doesn't support templates      //
//   define CL_NO_TEMPLATES below.                   //
//                                                   //
///////////////////////////////////////////////////////

#ifndef CL_NO_TEMPLATES
        #ifdef _MSC_VER
          #if  _MSC_VER <= 800
                //   Microsoft's C 7.0 and Visual C++
                //   8.0 do not support templates.
                #define CL_NO_TEMPLATES
          #endif
        #endif
//        #if YOUR_NO_TEMPLATES_CC
//                #define CL_NO_TEMPLATES
//        #endif
#endif

///////////////////////////////////////////////////////
//                                                   //
//   If your compiler doesn't support exceptions     //
//   define CL_NO_EXCEPTIONS below.                  //
//                                                   //
///////////////////////////////////////////////////////

#ifndef CL_NO_EXCEPTIONS
  #if defined(_MSC_VER)
    #if _MSC_VER <= 800
        //   Microsoft's C 7.0 and Visual C++ 8.0 do
        //   not support exceptions.
        #define CL_NO_EXCEPTIONS
    #endif
  #elif defined(__BCPLUSPLUS__)
    #if __BCPLUSPLUS__ < 0x0320
        //   Borland C++ compilers prior to 4.0 did 
        //   not support exceptions.
        #define CL_NO_EXCEPTIONS
    #endif
//#elif YOUR_NO_EXCEPTIONS_CC
//        #define CL_NO_EXCEPTIONS
  #endif
#endif // !defined(CL_NO_EXCEPTIONS)

///////////////////////////////////////////////////////
//                                                   //
//   To enable Container Lite methods to             //
//   throw the xalloc exception comment              //
//   out the define CL_NO_THROW_XALLOC below.        //
//   Define CL_XALLOC as the exception thrown.       //
//                                                   //
///////////////////////////////////////////////////////

#if !defined(CL_THROW_XALLOC)  \
 && !defined(CL_NO_THROW_XALLOC)  \
 && !defined(CL_NO_EXCEPTIONS)
        #define CL_NO_THROW_XALLOC
#endif
#define CL_XALLOC xalloc
#if defined(CL_NO_THROW_XALLOC)\
        && !defined(CL_NO_EXCEPTIONS)
        #define CL_try_new   try {
        // Flesh in catch (xalloc) {} as required.
        // CL primitives will return 0 on failures
        // as before unless you throw an additional
        // exception!
        #define CL_catch_xalloc(NULLD)           \
                } catch (CL_XALLOC)  { NULLD = 0; }
        // CL will return 0 if the catch block doesn't
        // terminate the program, even with
        // set_new_handler(0)!
#else
        #define CL_try_new
        #define CL_catch_xalloc(NULLD)
#endif

#ifndef limits_h
  #include <limits.h>        // UINT_MAX
  #ifndef limits_h
    #define limits_h
  #endif
#endif

#ifndef iostream_h
  #include <iostream.h>
  #ifndef iostream_h
    #define iostream_h
  #endif
#endif

#ifndef iomanip_h
  #include <iomanip.h>
  #ifndef iomanip_h
    #define iomanip_h
  #endif
#endif

#ifndef string_h
  #include <string.h>
  #ifndef string_h
    #define string_h
  #endif
#endif

#ifndef CL_NO_EXCEPTIONS
  #ifndef except_h
    #include <except.h>
    #ifndef except_h
      #define except_h
    #endif
  #endif
#endif

/*  stream delimiters/manipulators  */

inline ostream& endm(ostream& os)
{ return os << endl; }
inline istream& nextm(istream& is)
{ (void) is.get(); return is; }

/*  special functions for strings  */

extern char * CL_strdup(const char * D);
    // CL_strdup() uses new instead of malloc!
extern ostream&  CL_puts(ostream& os, const char * D);
extern istream&  CL_gets(istream& is, char *& D);

/*  cl compare pointer types  */

#define CLblank
#define CLcmP(ID,ITEM)  \
        int (*ID)(const ITEM * D1, const ITEM * D2)
#define CLkeyCmP(ID,ITEM) \
        int (*ID)(const void * K, const ITEM * D)
typedef CLcmP(voidCmP,void);
typedef CLkeyCmP(voidKeyCmP,void);
#define CLcmPcast(ID,ITEM)  \
        (CLcmP(CLblank,ITEM)) ID
#define CLkeyCmPcast(ID,ITEM) \
        (CLkeyCmP(CLblank,ITEM)) ID
#define CLKCC(ITEM) (CLkeyCmP(CLblank,ITEM))
#define CLKCCSM(ITEM,CMP) (CLKCC(ITEM) ITEM :: CMP)
#define CLcmP0(ITEM)  CLcmPcast(0,ITEM)
#define CLkeyCmP0(ITEM) CLkeyCmPcast(0,ITEM)
#define voidCmP0 ((voidCmP)0)
#define voidKeyCmP0 ((voidKeyCmP)0)
#define CmPfnc cmP() const

#define CL_HONOR_SORTED 1

// uniquenessGimmick

typedef struct { int x; } * defaultConstructor;
#define defaultConstruct  ((defaultConstructor)0)

/*  cl constants  */

#define CL_MAXNODES     ((unsigned) \
                        (UINT_MAX/sizeof(void *)))
#define CL_LIMIT        20U
#define CL_DELTA        10U
#define CL_NOTFOUND     CL_MAXNODES
#define CL_SORTED       0x0001U
#define CL_BIND_ONLY    0x0000U
#define CL_ASG          0x0002U
#define CL_READ         0x0004U
#define CL_READ_DEL     0x0008U
#define CL_NEW          0x0010U
#define CL_NEWED        0x0020U
#define CL_DEL          0x0040U
#define CL_SAVE         0x0080U
#define CL_ALL_FLAGS    0x00FFU
#define CL_ANDS         (CL_ASG|CL_NEW|CL_DEL|CL_SAVE)

// avoid [0U] ambiguity with [KeyPtr]

#define CL_0U           -0

#define CL_forEach(CL)  \
    for ((void)(CL).setCurNode(0); (CL).get(); ++(CL))

class cl
{

  public:

        enum RelCmp { EQ, NE, GT, GE, LT, LE };

  private:

        void * nD, ** vL;   // pointers to danglers
        static char * vD;
        unsigned lowLimit, lowThreshold, first;
        void **  linkS;
        unsigned _limit,    _delta,   _nodes;
        unsigned _maxNodes, _flags;
        voidCmP  _cmP;
        static int  rcmp(const void * K, const void * D,
                voidKeyCmP kcmP, cl::RelCmp rc);
        void init(unsigned flags = CL_BIND_ONLY,
                unsigned maxNodes = CL_MAXNODES,
                unsigned limit = CL_LIMIT,
                unsigned delta = CL_DELTA);
        void setND(void * ND = 0)
                { if (nD) deleteD(nD); nD = ND; }
        void resetND()  { nD = 0; }
        void setVL(void ** VL = 0)
                { if (vL) delete[] vL; vL = VL; }
        void resetVL()  { vL = 0; }
        friend istream& CL_gets
                (istream& is, char *& D);
          static void setVD(char * VD = 0)
                { if (vD) delete[] vD; vD = VD; }
          static void resetVD()  { vD = 0; }
        void * cloneD(const void * D);
        void * decloneD(void * D);
        void * deleteLeak(void * OK, void * D );

  protected:

        unsigned _curNode;

        cl  (defaultConstructor)  { init(); }
        void    assign(const cl& b, int deep = 1);
        void    destruct();
        virtual voidCmP cmPD() const
                        { return 0; }
        virtual void * assignD(void *, const void *)
                        { return (void *)0; }
        virtual void * newD(const void *)
                        { return (void *)0; }
        virtual void   deleteD(void * D)
                        { delete D; }
        virtual int    attachD(void *)
                        { return 1; }
        virtual void   detachD(void *)
                        {}
        virtual int    putD(ostream&, void *)
                        { return 0; }
        virtual void * getD(istream&)
                        { return (void *)0; }
        virtual int    putH(ostream&)
                        { return 1; }
        virtual int    getH(istream& is)
                        { return 1; }
        virtual int    put(ostream& os);
        virtual int    get(istream& is);
        int     load (const char * filename);
        int     save (const char * filename);
        friend  ostream& operator<<
                        (ostream& os, cl& b);
        friend  istream& operator>>
                        (istream& is, cl& b);

public:

        // For use by attach(), detach()
        virtual void * thisCL() { return this; }

/*  Constructors and destructor  */

        cl     (unsigned flags_ = CL_BIND_ONLY,
                unsigned maxNodes_ = CL_MAXNODES,
                unsigned limit_ = CL_LIMIT,
                unsigned delta_ = CL_DELTA)
                { init(flags_,maxNodes_,limit_,delta_); }
        cl     (void ** argv, unsigned argc = 0U,
                unsigned flags_ = CL_BIND_ONLY);
        ostream& storeOn(ostream& os)
                { if (flags(CL_SAVE)) (void) put(os);
                  return os; }
        istream& readFrom(istream& is)
                { (void) get(is); return is; }
        void ** vector(void ** argv = (void **)0,
                        unsigned argc = 0U) const;
        cl&     operator=(const cl& b)
                { assign(b); return *this; }
        int     cmp(const cl& b) const;
        virtual ~cl()  { destruct(); }
/*
        You must override this destructor in any
        descendant that overrides deleteD() or
        detachD(), i.e.
        virtual ~DerivedFromCL()
                { cl::destruct(); }
        This is because base destructors are called
        after vfts are reset to their default values
        for the base level.  This means that
        cl::~cl() calls cl::detachD()
        and perhaps cl::deleteD() instead of
        the intended DerivedFromCL::detachD()
        or DerivedFromCL::deleteD().
*/

/*  Housekeeping Primitives  */

        void catch_reset()
                { setND(); setVL(); setVD(); }
        unsigned limit() const { return _limit; }
        unsigned setLimit(unsigned newLimit);
        unsigned pack()  { return setLimit(_nodes); }
        unsigned delta() const { return _delta; }
        unsigned setDelta(unsigned newDelta
                        = CL_DELTA);
        unsigned nodes() const { return _nodes; }
        unsigned maxNodes() const { return _maxNodes; }
        unsigned setMaxNodes(unsigned newMaxNodes
                        = CL_MAXNODES);
        unsigned vacancy() const
                        { return _maxNodes - _nodes; }
        unsigned vacancyNonElastic() const
                        { return _limit - _nodes; }
        unsigned flags(unsigned flags_ = CL_ALL_FLAGS)
                        const
                    { return (_flags & flags_); }
        void     setFlags(unsigned flags_)
                    { _flags |= flags_; }
        void     resetFlags(unsigned flags_)
                    { _flags &= ~flags_; }

/*  Elastic Array Primitives  */

        void *   atIns(unsigned n, void * D);
        void *   atInsNew(unsigned n, const void * D)
                     { return decloneD
                       (atIns(n,cloneD(D))); }
        void *   atInsDL(unsigned n, void * D)
                     { return deleteLeak
                       (atIns(n,D),D); }
        void *   atRmv(unsigned n);
        void     allRmv();
        int      atDel(unsigned n);
        void *   atDelAsg(unsigned n, void * D)
                     { return (atGetAsg(n,D)?
                       (atDel(n), D) : 0); }
        int      allDel();
        int      atClr(unsigned n)
                     { return (atDel(n)?
                       1 : (atRmv(n) != 0)); }
        void     allClr()  { destruct(); }
        void *   atPut(unsigned n, void * D);
        void *   atPutNew(unsigned n, const void * D);
        void *   atPutDL(unsigned n, void * D)
                     { return deleteLeak
                       (atPut(n,D),D); }
        void *   atPutAsg(unsigned n, const void * D);
        void *   atGet(unsigned n) const;
        void *   operator[](unsigned n) const
                        { return atGet(n); }
        void *   atGetAsg(unsigned n, void * D);
        void *   atXchg(unsigned n, void * D);
        unsigned indexOf(const void * D) const;

/*  Stack - Deque - Queue Primitives  */

        void *   push(void * D)
                        { return atIns(0U,D); }
        void *   pushNew(const void * D)
                        { return atInsNew(0U,D); }
        void *   pushDL(void * D)
                     { return atInsDL(0U,D); }
        void *   pop()  { return atRmv(0U); }
        int      popDel()  { return atDel(0U); }
        void *   popDelAsg(void * D)
                        { return atDelAsg(0U,D); }
        void *   top() const { return atGet(0U); }
        void *   topAsg(void * D)
                        { return atGetAsg(0U,D); }
        void *   insQ(void * D)
                        { return atIns(_nodes,D); }
        void *   insQNew(const void * D)
                        { return atInsNew(_nodes,D); }
        void *   insQDL(void * D)
                        { return atInsDL(_nodes,D); }
        void *   unQ()  { return atRmv(_nodes-1); }
        int      unQDel()  { return atDel(_nodes-1); }
        void *   unQDelAsg(void * D)
                        { return atDelAsg(_nodes-1,D);}
        void *   rear() const
                        { return atGet(_nodes-1); }
        void *   rearAsg(void * D)
                        { return atGetAsg(_nodes-1,D);}

/*  List (single and double linked) Primitives  */

        unsigned curNode() const
                 { return ((_curNode < _nodes)?
                   _curNode : CL_NOTFOUND); }
        void *   setCurNode(unsigned n = CL_NOTFOUND);
        void *   insBefore(void * D);
        void *   insBeforeNew(const void * D)
                     { return decloneD
                       (insBefore(cloneD(D))); }
        void *   insBeforeDL(void * D)
                     { return deleteLeak
                       (insBefore(D),D); }
        void *   rmvStayPut();
        int      delStayPut();
        void *   delAsgStayPut(void * D)
                     { return (atGetAsg(_curNode,D)?
                       (delStayPut(), D) : 0); }
        int      clrStayPut();
        void *   insAfter(void * D);
        void *   insAfterNew(const void * D)
                     { return decloneD
                       (insAfter(cloneD(D))); }
        void *   insAfterDL(void * D)
                     { return deleteLeak
                       (insAfter(D),D); }
        void *   rmvBackup();
        int      delBackup();
        void *   delAsgBackup(void * D)
                     { return (atGetAsg(_curNode,D)?
                       (delBackup(), D) : 0); }
        int      clrBackup();
        void *   put(void * D)
                        { return atPut(_curNode,D); }
        void *   putNew(const void * D)
                        { return atPutNew(_curNode,D); }
        void *   putDL(void * D)
                        { return atPutDL(_curNode,D); }
        void *   putAsg(const void * D)
                        { return atPutAsg(_curNode,D); }
        void *   get() const { return atGet(_curNode); }
        void *   getAsg(void * D)
                        { return atGetAsg(_curNode,D); }
        void *   next();
        void *   operator++()  { return next(); }
        void *   nextAsg(void * D);
        void *   prev();
        void *   operator--()  { return prev(); }
        void *   prevAsg(void * D);

/* Priority Q, Set, Bag, Dictionary, Sort Primitives */

        unsigned sorted() const
                    { return (_flags & CL_SORTED); }
        void     unSort()  { _flags &= ~CL_SORTED; }
        void     setCmP(voidCmP cmP_ = voidCmP0,
                     int resetSortFlag = 1)
                     { _cmP = (cmP_? cmP_ : cmPD());
                       if (resetSortFlag && _nodes)
                         _flags &= ~CL_SORTED; }
        voidCmP  cmP() const { return _cmP; }
        int      sort(voidCmP cmP_ = voidCmP0);
        void *   insSort(void * D);
        void *   insSortNew(const void * D)
                     { return decloneD
                       (insSort(cloneD(D))); }
        void *   insSortDL(void * D)
                     { return deleteLeak
                       (insSort(D),D); }
        void *   insUnique(void * D);
        void *   insUniqueNew(const void * D);
        void *   insUniqueDL(void * D)
                     { return deleteLeak
                       (insUnique(D),D); }
        void *   findFirst(const void * K,
                        unsigned& idx,
                        voidKeyCmP kcmP = 0,
                        int honorSortFlag = 1,
                        cl::RelCmp rc = EQ) const;
        void *   findNext(const void * K,
                        unsigned& idx,
                        voidKeyCmP kcmP = 0,
                        int honorSortFlag = 1,
                        cl::RelCmp rc = EQ) const;
        void *   findFirst(const void * K,
                        voidKeyCmP kcmP = 0,
                        int honorSortFlag = 1,
                        cl::RelCmp rc = EQ)
                        { return findFirst(K,_curNode,
                          kcmP,honorSortFlag,rc); }
        void *   findNext(const void * K,
                        voidKeyCmP kcmP = 0,
                        int honorSortFlag = 1,
                        cl::RelCmp rc = EQ)
                        { return findNext(K,_curNode,
                          kcmP,honorSortFlag,rc); }

};      /*  class cl  */

inline ostream& operator<<(ostream& os, cl& b)
{ return b.storeOn(os); }

inline istream& operator>>(istream& is, cl& b)
{ return b.readFrom(is); }

typedef void (*GenericFnC)(void);
#define GenericFnC0  ((GenericFnC)0)

class FunctionRegistry : private cl
{

public:

        FunctionRegistry() : cl(CL_DEL) {}
        int regFnC(GenericFnC fnC, unsigned id);
        unsigned fnC_2_ID(GenericFnC fnC);
        GenericFnC ID_2_fnC(unsigned id);
        void forget() { (void) allDel(); }
        ~FunctionRegistry() {}

};

extern  FunctionRegistry fnCv;

// Do not invoke the macros below in global instances
// member functions of any class because fnCv may not
// yet be constructed!

#define Forget_Functions() fnCv.forget()
#define Register_Function(fnC,id)  \
                fnCv.regFnC((GenericFnC)fnC,id)
#define Register_CmP(cmP,id)  \
                Register_Function(cmP,id)
#define fnC2ID(fnC) fnCv.fnC_2_ID((GenericFnC)fnC)
#define ID2fnC(id)  fnCv.ID_2_fnC(id)
#define cmP2ID(cmP) fnC2ID(cmP)
#define ID2cmP(id) ((voidCmP)ID2fnC(id))

/*  CL's C++ TEMPLATES  */

#ifndef CL_NO_TEMPLATES

#ifndef assert_h
  #include <assert.h>
  #ifndef assert_h
    #define assert_h
  #endif
#endif

/*  ITEM descriptors  */

inline int CL_cmp(const int& i1, const int& i2)
    { return i1 - i2; }
inline int CL_cmp(const long& i1, const long& i2)
    { return i1 - i2; }
inline int CL_cmp(const short& i1, const short& i2)
    { return i1 - i2; }
inline int CL_cmp
    (const unsigned& i1, const unsigned& i2)
    { return ((i1 > i2)? 1 : ((i1 == i2)? 0 : -1)); }
inline int CL_cmp
    (const unsigned long & i1, const unsigned long& i2)
    { return ((i1 > i2)? 1 : ((i1 == i2)? 0 : -1)); }
inline int CL_cmp(const unsigned short& i1,
    const unsigned short& i2)
    { return ((i1 > i2)? 1 : ((i1 == i2)? 0 : -1)); }
inline int CL_cmp(const float& i1, const float& i2)
    { return ((i1 > i2)? 1 : ((i1 == i2)? 0 : -1)); }
inline int CL_cmp(const double& i1, const double& i2)
    { return ((i1 > i2)? 1 : ((i1 == i2)? 0 : -1)); }
inline int CL_cmp(const long double& i1,
    const long double& i2)
    { return ((i1 > i2)? 1 : ((i1 == i2)? 0 : -1)); }

template <class ITEM>
inline void * CL_assign
    (ITEM& i1, const ITEM& i2, unsigned /* flags */)
    { i1 = i2; return &i1; }
template <class ITEM>
inline void * CL_newCopy(const ITEM& i)
    { return new ITEM(i); }
template <class ITEM>
inline void CL_delete(ITEM * I)
    { delete I; }
template <class ITEM>
inline int CL_attach(ITEM&, cl *)
    { return 1; }
template <class ITEM>
inline void CL_detach(ITEM&, cl *)
    {}
template <class ITEM>
inline int CL_cmp(const ITEM& i1, const ITEM& i2)
    { return i1.cmp(i2); }
template <class ITEM>
inline int CL_put(ostream& os, ITEM& i)
    { return ((os << i << endm)? 1 : 0); }
template <class ITEM>
inline void * CL_get(istream& is, ITEM * /* gimmick */)
    {
        ITEM * I = new ITEM;
        if (I) if (!(is >> *I >> nextm))
            { delete I; I = 0; }
        return I;
    }

#define CL_NO_assign(ITEM) \
    inline void * CL_assign \
    (ITEM&, const ITEM&, unsigned) \
    { return 0; }
#define CL_NO_newCopy(ITEM) \
    inline void * CL_newCopy(const ITEM&) \
    { return 0; }
#define CL_NO_delete(ITEM) \
    inline void CL_delete(ITEM *) \
    { assert(0); }
#define CL_NO_cmp(ITEM) \
    inline int CL_cmp(const ITEM&, const ITEM&) \
    { assert(0); return 0; }
#define CL_NO_stream(ITEM) \
    inline int CL_put(ostream&, ITEM&) \
    { return 0; } \
    inline void * CL_get(istream&, ITEM *) \
    { return 0; }

template <class ITEM>
struct CL_traits
{
    static void * assign(ITEM& i1, const ITEM& i2,
        unsigned flags)
        { return CL_assign(i1,i2,flags); }
    static void * newCopy(const ITEM & i)
        { return CL_newCopy(i); }
    static void _delete(ITEM * I) { CL_delete(I); }
    static int attach(ITEM& i, cl * C)
        { return CL_attach(i,C); }
    static void detach(ITEM& i, cl * C)
        { CL_detach(i,C); }
    static int cmp(const ITEM& i1, const ITEM& i2)
        { return CL_cmp(i1,i2); }
    static int put(ostream& os, ITEM& i)
        { return CL_put(os,i); }
    static void * get(istream& is)
        { return CL_get(is,(ITEM *)0); }
};

struct CL_traits<char>
{
    static void * assign(char&, const char&, unsigned)
        {  return 0; }
    static void * newCopy(const char& i)
        { return CL_strdup(&i); }
    static void _delete(char * I) { delete[] I; }
    static int attach(char&, cl *)
        { return 1; }
    static void detach(char&, cl *) {}
    static int cmp(const char& i1, const char& i2)
        { return strcmp(&i1,&i2); }
    static int put(ostream& os, char& i)
        { return (CL_puts(os,&i)? 1 : 0); }
    static char * get(istream& is)
        { char * I = 0; (void) CL_gets(is,I);
          return I; }
};

#define CL_StrongTyping(ITEM) \
    struct CL_traits<ITEM > \
    { \
      static void * assign(ITEM&, \
        const ITEM&, unsigned) \
        { return 0; } \
      static void * newCopy(const ITEM & i) \
        { return 0; } \
      static void _delete(ITEM * I) \
        { delete I; } \
      static int attach(ITEM&, cl *) \
        { return 1; } \
      static void detach(ITEM&, cl *) \
        { return; } \
      static int cmp(const ITEM&, \
        const ITEM&) \
        { assert(0); return 0; } \
      static int put(ostream&, ITEM&) \
        { return 0; } \
      static void * get(istream&) \
        { return 0; } \
    }

#define CL_StrongTypingCmp(ITEM) \
    struct CL_traits<ITEM > \
    { \
      static void * assign(ITEM&, \
        const ITEM&, unsigned) \
        { return 0; } \
      static void * newCopy(const ITEM & i) \
        { return 0; } \
      static void _delete(ITEM * I) \
        { delete I; } \
      static int attach(ITEM&, cl *) \
        { return 1; } \
      static void detach(ITEM&, cl *) \
        { return; } \
      static int cmp(const ITEM& i1, \
        const ITEM& i2) \
        { return CL_cmp(i1,i2); } \
      static int put(ostream&, ITEM&) \
        { return 0; } \
      static void * get(istream&) \
        { return 0; } \
    }

#define CL_StrongTypingStream(ITEM) \
    struct CL_traits<ITEM > \
    { \
      static void * assign(ITEM&, \
        const ITEM&, unsigned) \
        { return 0; } \
      static void * newCopy(const ITEM & i) \
        { return 0; } \
      static void _delete(ITEM * I) \
        { delete I; } \
      static int attach(ITEM&, cl *) \
        { return 1; } \
      static void detach(ITEM&, cl *) \
        { return; } \
      static int cmp(const ITEM&, \
        const ITEM&) \
        { assert(0); return 0; } \
      static int put(ostream& os, ITEM& i) \
        { return CL_put(os,i); } \
      static void * get(istream& is) \
        { return CL_get(is,(ITEM *)0); } \
    }

#define CL_StrongTypingCmpStream(ITEM) \
    struct CL_traits<ITEM > \
    { \
      static void * assign(ITEM&, \
        const ITEM&, unsigned) \
        { return 0; } \
      static void * newCopy(const ITEM & i) \
        { return 0; } \
      static void _delete(ITEM * I) \
        { delete I; } \
      static int attach(ITEM&, cl *) \
        { return 1; } \
      static void detach(ITEM&, cl *) \
        { return; } \
      static int cmp(const ITEM& i1, \
        const ITEM& i2) \
        { return CL_cmp(i1,i2); } \
      static int put(ostream& os, ITEM& i) \
        { return CL_put(os,i); } \
      static void * get(istream& is) \
        { return CL_get(is,(ITEM *)0); } \
    }

/*
    // Template to define you own traits
    struct CL_traits<ITEM >
    {
      static void * assign(ITEM&,
        const ITEM&, unsigned)
        { return 0; }
      static void * newCopy(const ITEM & i)
        { return 0; }
      static void _delete(ITEM * I)
        { delete I; }
      static int attach(ITEM&, cl *)
        { return 1; }
      static void detach(ITEM&, cl *)
        { return; }
      static int cmp(const ITEM&,
        const ITEM&)
        { assert(0); return 0; }
      static int put(ostream&, ITEM&)
        { return 0; }
      static void * get(istream&)
        { return 0; }
    };
*/

template <class ITEM>
class CL : public cl
{

protected:

        CL  (defaultConstructor)
                        : cl(defaultConstruct) {}
        void assign(const CL<ITEM>& b, int deep = 1)
                        { cl::assign(b,deep); }
        // cl::  destruct;
        virtual  voidCmP cmPD() const
                        { return ((voidCmP)
                        CL_traits<ITEM>::cmp); }
        virtual  void * assignD(void * D,
                        const void * S)
                        { return CL_traits<ITEM>::
                        assign(*(ITEM *)D,
                        *(const ITEM *)S,
                        cl::flags()); }
        virtual  void * newD(const void * D)
                        { return  CL_traits<ITEM>::
                        newCopy(*(const ITEM *)D); }
        virtual  void   deleteD(void * D)
                        { CL_traits<ITEM>::
                        _delete((ITEM *)D); }
        virtual  int    attachD(void * D)
                        { return CL_traits<ITEM>::
                        attach(*(ITEM *)D,this); }
        virtual  void   detachD(void * D)
                        { CL_traits<ITEM>::
                        detach(*(ITEM *)D,this); }
        virtual  int    putD(ostream& os, void * D)
                        { return CL_traits<ITEM>::
                        put(os,*(ITEM *)D); }
        virtual  void * getD(istream& is)
                        { return CL_traits<ITEM>::
                        get(is); }
        virtual  int    put(ostream& os)
                        { return cl::put(os); }
        virtual  int    get(istream& is)
                        { return cl::get(is); }

public:

/*  Constructors and destructor  */

        CL (unsigned flags = CL_BIND_ONLY,
                unsigned maxNodes = CL_MAXNODES,
                unsigned limit = CL_LIMIT,
                unsigned delta = CL_DELTA) :
                cl(flags,maxNodes,limit,delta)
                {}
        CL (ITEM * argv[],
                unsigned argc = 0U,
                unsigned flags = CL_BIND_ONLY)
                : cl((void **)argv,argc,flags)
                {}
        CL (const CL<ITEM>& b)
                : cl(defaultConstruct)
                { assign(b); }
        CL (const char * filename)
                : cl(defaultConstruct)
                { (void) cl::load(filename); }
        static CL<ITEM> * newInstance
                (const char * filename)
                { return new CL<ITEM>(filename); }
        CL (istream& is)
                : cl(defaultConstruct)
                { (void) cl::readFrom(is); }
        static CL<ITEM> * newInstance(istream& is)
                { return new CL<ITEM>(is); }
        int     load(const char * filename)
                { return cl::load(filename); }
        int     save(const char * filename)
                { return cl::save(filename); }
        //cl::  storeOn;
        //cl::  readFrom;
        CL<ITEM>& operator=(const CL<ITEM>& b)
                { assign(b); return *this; }
        int     cmp(const CL<ITEM>& b) const
                { return cl::cmp(b); }
        ITEM ** vector(ITEM ** argv = (ITEM **)0,
                unsigned argc = 0U) const
                { return (ITEM **)cl::
                vector((void **)argv,argc); }
        virtual ~CL() { destruct(); }

/*  Housekeeping Primitives  */

/*
        cl::    limit;
        cl::    setLimit;
        cl::    pack;
        cl::    delta;
        cl::    setDelta;
        cl::    nodes;
        cl::    maxNodes;
        cl::    setMaxNodes;
        cl::    vacancy;
        cl::    vacancyNonElastic;
        cl::    flags;
        cl::    setFlags;
        cl::    resetFlags;
*/

/*  Elastic Array Primitives  */

        ITEM *   atIns(unsigned n, ITEM * D)
                        { return (ITEM *)cl::
                        atIns(n,(void *)D); }
        ITEM *   atInsNew(unsigned n, const ITEM * D)
                        { return (ITEM *)cl::
                        atInsNew(n,(const void *)D); }
        ITEM *   atInsDL(unsigned n, ITEM * D)
                        { return (ITEM *)cl::
                        atInsDL(n,(void *)D); }
        ITEM *   atRmv(unsigned n)
                        { return (ITEM *)cl::
                        atRmv(n); }
        //cl::   allRmv;
        //cl::   atDel;
        ITEM *   atDelAsg(unsigned n, ITEM * D)
                        { return (ITEM *)cl::
                        atDelAsg(n,(void *)D); }
        //cl::   allDel;
        //cl::   atClr;
        //cl::   allClr;
        ITEM *   atPut(unsigned n, ITEM * D)
                        { return (ITEM *)cl::
                        atPut(n,(void *)D); }
        ITEM *   atPutNew(unsigned n, const ITEM * D)
                        { return (ITEM *)cl::
                        atPutNew(n,(const void *)D); }
        ITEM *   atPutDL(unsigned n, ITEM * D)
                        { return (ITEM *)cl::
                        atPutDL(n,(void *)D); }
        ITEM *   atPutAsg(unsigned n, const ITEM * D)
                        { return (ITEM *)cl::
                        atPutAsg(n,(const void *)D); }
        ITEM *   atGet(unsigned n) const
                        { return (ITEM *)cl::
                        atGet(n); }
        ITEM *   operator[](unsigned n) const
                        { return atGet(n); }
        ITEM *   atGetAsg(unsigned n, ITEM * D)
                        { return (ITEM *)cl::
                        atGetAsg(n,(void *) D); }
        ITEM *   atXchg(unsigned n, ITEM * D)
                        { return (ITEM *)cl::
                        atXchg(n,(void *) D); }
        unsigned indexOf(const ITEM * D) const
                        { return cl::
                        indexOf((const void *)D); }

/*  Stack - Deque - Queue Primitives  */

        ITEM *   push(ITEM * D)
                        { return (ITEM *)cl::
                        push((void *) D); }
        ITEM *   pushNew(const ITEM * D)
                        { return (ITEM *)cl::
                        pushNew((const void *)D); }
        ITEM *   pushDL(ITEM * D)
                        { return (ITEM *)cl::
                        pushDL((void *) D); }
        ITEM *   pop()
                        { return (ITEM *)cl::
                        pop(); }
        //cl::   popDel;
        ITEM *   popDelAsg(ITEM * D)
                        { return (ITEM *)cl::
                        popDelAsg((void *)D); }
        ITEM *   top() const
                        { return (ITEM *)cl::
                                top(); }
        ITEM *   topAsg(ITEM * D)
                        { return (ITEM *)cl::
                        topAsg((void *)D); }
        ITEM *   insQ(ITEM * D)
                        { return (ITEM *)cl::
                        insQ((void *)D); }
        ITEM *   insQNew(const ITEM * D)
                        { return (ITEM *)cl::
                        insQNew((const void *)D); }
        ITEM *   insQDL(ITEM * D)
                        { return (ITEM *)cl::
                        insQDL((void *)D); }
        ITEM *   unQ()
                        { return (ITEM *)cl::
                        unQ(); }
        //cl::   unQDel;
        ITEM *   unQDelAsg(ITEM * D)
                        { return (ITEM *)cl::
                        unQDelAsg((void *)D); }
        ITEM *   rear() const
                        { return (ITEM *)cl::
                                rear(); }
        ITEM *   rearAsg(ITEM * D)
                        { return (ITEM *)cl::
                        rearAsg((void *)D); }

/*  List (single and double linked) Primitives  */

        //cl::   curNode;
        ITEM *   setCurNode(unsigned n = CL_NOTFOUND)
                        { return (ITEM *)cl::
                        setCurNode(n); }
        ITEM *   insBefore(ITEM * D)
                        { return (ITEM *)cl::
                        insBefore((void *)D); }
        ITEM *   insBeforeNew(const ITEM * D)
                        { return (ITEM *)cl::
                        insBeforeNew((const void *)D); }
        ITEM *   insBeforeDL(ITEM * D)
                        { return (ITEM *)cl::
                        insBeforeDL((void *)D); }
        ITEM *   rmvStayPut()
                        { return (ITEM *)cl::
                        rmvStayPut(); }
        //cl::   delStayPut;
        ITEM *   delAsgStayPut(ITEM * D)
                        { return (ITEM *)cl::
                        delAsgStayPut((void *)D); }
        //cl::   clr;
        ITEM *   insAfter(ITEM * D)
                     { return (ITEM *)cl::
                       insAfter((void *)D); }
        ITEM *   insAfterNew(const ITEM * D)
                     { return (ITEM *)cl::
                       insAfterNew((const void *)D); }
        ITEM *   insAfterDL(ITEM * D)
                        { return (ITEM *)cl::
                        insAfterDL((void *)D); }
        ITEM *   rmvBackup()
                     { return (ITEM *)cl::
                       rmvBackup(); }
        //cl::   delBackup;
        ITEM *   delAsgBackup(ITEM * D)
                     { return (ITEM *)cl::
                       delAsgBackup((void *)D); }
        //cl::   clrBackup;
        ITEM *   put(ITEM * D)
                        { return (ITEM *)cl::
                        put((void *)D); }
        ITEM *   putNew(const ITEM * D)
                        { return (ITEM *)cl::
                        putNew((const void *)D); }
        ITEM *   putDL(ITEM * D)
                        { return (ITEM *)cl::
                        putDL((void *)D); }
        ITEM *   putAsg(const ITEM * D)
                        { return (ITEM *)cl::
                        putAsg((const void *)D); }
        ITEM *   get() const
                        { return (ITEM *)cl::
                        get(); }
        ITEM *   getAsg(ITEM * D)
                        { return (ITEM *)cl::
                        getAsg((void *)D); }
        ITEM *   next()
                        { return (ITEM *)cl::
                        next(); }
        ITEM *   operator++()
                        { return next(); }
        ITEM *   nextAsg(ITEM * D)
                        { return (ITEM *)cl::
                        nextAsg((void *)D); }
        ITEM *   prev()
                        { return (ITEM *)cl::
                        prev(); }
        ITEM *   operator--()
                        { return prev(); }
        ITEM *   prevAsg(ITEM * D)
                        { return (ITEM *)cl::
                        prevAsg((void *)D); }

/* Priority Q, Set, Bag, Dictionary, Sort Primitives */

        //cl::   sorted;
        //cl::   unSort;
        void     setCmP(CLcmP(cmP_,ITEM)
                     = CLcmP0(ITEM),
                     int resetSortFlag = 1)
                     { cl::setCmP
                      ((voidCmP)cmP_,resetSortFlag); }
        CLcmP(CmPfnc,ITEM) { return CLcmPcast
                        (cl::cmP(),ITEM); }
        int      sort(CLcmP(cmP_,ITEM)
                        = CLcmP0(ITEM))
                        { return cl::sort
                        ((voidCmP)cmP_); }
        ITEM *   insSort(ITEM * D)
                        { return (ITEM *)cl::
                        insSort((void *)D); }
        ITEM *   insSortNew(const ITEM * D)
                        { return (ITEM *)cl::
                        insSortNew((const void *)D); }
        ITEM *   insSortDL(ITEM * D)
                        { return (ITEM *)cl::
                        insSortDL((void *)D); }
        ITEM *   insUnique(ITEM * D)
                        { return (ITEM *)cl::
                        insUnique((void *)D); }
        ITEM *   insUniqueNew(const ITEM * D)
                        { return (ITEM *)cl::
                        insUniqueNew((const void *)D);}
        ITEM *   insUniqueDL(ITEM * D)
                        { return (ITEM *)cl::
                        insUniqueDL((void *)D); }
        ITEM *   findFirst(const void * K,
                        unsigned& idx,
                        CLkeyCmP(kcmP,ITEM)
                          = CLkeyCmP0(ITEM),
                        int honorSortFlag = 1,
                        cl::RelCmp rc = EQ) const
                        { return (ITEM *) cl::
                        findFirst(K,idx,
                        voidKeyCmP(kcmP),
                        honorSortFlag,rc); }
        ITEM *   findNext(const void * K,
                        unsigned& idx,
                        CLkeyCmP(kcmP,ITEM)
                          = CLkeyCmP0(ITEM),
                        int honorSortFlag = 1,
                        cl::RelCmp rc = EQ) const
                        { return (ITEM *) cl::
                        findNext(K,idx,
                        voidKeyCmP(kcmP),
                        honorSortFlag,rc); }
        ITEM *   findFirst(const void * K,
                        CLkeyCmP(kcmP,ITEM)
                          = CLkeyCmP0(ITEM),
                        int honorSortFlag = 1,
                        cl::RelCmp rc = EQ)
                        { return findFirst(K,_curNode,
                          kcmP,honorSortFlag,rc); }
        ITEM *   operator[](const void * K)
                        { return (ITEM *)cl::
                        findFirst(K); }
        ITEM *   findNext(const void * K,
                        CLkeyCmP(kcmP,ITEM)
                          = CLkeyCmP0(ITEM),
                        int honorSortFlag = 1,
                        cl::RelCmp rc = EQ)
                        { return findNext(K,_curNode,
                          kcmP,honorSortFlag,rc); }

};      /*  class CL  */

template <class ITEM>
class CL_iterator
{
    CL<ITEM > * B;
    unsigned _idx;
  public:
    CL_iterator() : B(0), _idx(CL_NOTFOUND) {}
    CL_iterator(CL<ITEM >& b)
        : B(&b), _idx(CL_NOTFOUND) {}
    CL_iterator<ITEM >& operator()(CL<ITEM >& b)
        { B = &b; _idx := CL_NOTFOUND; return *this; }
    operator ITEM *()
        { return (B? B->atGet(_idx) : 0); }
    ITEM * operator++()
        { return (B? B->atGet((++_idx > B->nodes())?
          (_idx = 0) : _idx) : 0); }
    ITEM * operator--()
        { return (B? B->atGet((_idx--)?
          _idx  : (_idx = B->nodes())) : 0); }
};

// Smalltalk like enumerators

template <class ITEM>
class CL_doBlock
{
    void (*__doIt)(ITEM& i);
  protected:
    friend CL_do<ITEM >;
    virtual void doIt(ITEM& i)
        { assert(__doIt); (*__doIt)(i); }
  public:
    CL_doBlock(void (*_doIt)(ITEM& i) = 0)
        : __doIt(_doIt) {}
};

template <class ITEM>
class CL_do
{
  public:
    CL_do<ITEM >& doIt(CL<ITEM >& b,
        CL_doBlock<ITEM >& d);
    CL_do<ITEM >& doIt(CL<ITEM >& b,
        void (*_doIt)(ITEM& i))
        { return doIt(b,CL_doBlock<ITEM >(_doIt)); }
    CL_do() {}
    CL_do(CL<ITEM >& b, CL_doBlock<ITEM >& d)
        { (void) doIt(b,d); }
    CL_do(CL<ITEM >& b, void (*_doIt)(ITEM& i))
        { (void) doIt(b,_doIt); }
};

template <class ITEM>
CL_do<ITEM >& CL_do<ITEM >::
    doIt(CL<ITEM >& b, CL_doBlock<ITEM >& d)
{
    for (unsigned i = 0; i < b.nodes(); i++)
        d.doIt(*b[i]);
    return *this;
}

enum CL_select_ops
    { CL_SELECT, CL_EXTRACT, CL_DISCARD, CL_CLONE };

template <class ITEM>
class CL_selectBlock
{
    int (*__select)(const ITEM& i);
  protected:
    friend CL_select<ITEM >;
    virtual int select(const ITEM& i)
        { assert(__select); return (*__select)(i); }
  public:
    CL_selectBlock(int (*_select)(const ITEM& i) = 0)
        : __select(_select) {}
};

template <class ITEM>
class CL_select : public CL<ITEM >
{
    int sense(int x, int reject)
        { return (reject? !x : x); }
  public:
    CL_select<ITEM >& select
       (CL<ITEM >& b, CL_selectBlock<ITEM >& s,
        CL_select_ops op = CL_SELECT, int reject = 0);
    CL_select<ITEM >& select(CL<ITEM >& b,
        int (*_select)(const ITEM& i),
        CL_select_ops op = CL_SELECT, int reject = 0)
        { return select(b,
          CL_selectBlock<ITEM >(_select),op,reject); }
    CL_select() : CL<ITEM >() {}
    CL_select(CL<ITEM >& b,
        CL_selectBlock<ITEM >& s,
        CL_select_ops op = CL_SELECT, int reject = 0)
        : CL<ITEM >() { (void) select(b,s,op,reject); }
    CL_select(CL<ITEM >& b,
        int (*_select)(const ITEM& i),
        CL_select_ops op = CL_SELECT, int reject = 0)
        : CL<ITEM >()
        { (void) select(b,_select,op,reject); }
};

template <class ITEM>
CL_select<ITEM >& CL_select<ITEM >::select
    (CL<ITEM >& b, CL_selectBlock<ITEM >& s,
    CL_select_ops op, int reject)
{
    unsigned i;
    allClr();
    switch (op)  {
        case CL_SELECT:
          setFlags(CL_BIND_ONLY);
          for (i = 0; i < b.nodes(); i++)
            if (sense(s.select(*b[i]),reject))
              (void) insQ(b[i]);
          break;
        case CL_EXTRACT:
          setFlags(b.flags());
          for (i = 0; i < b.nodes(); )
            if (sense(s.select(*b[i]),reject))
              if (flags(CL_DEL))
                (void) insQDL(b.atRmv(i));
              else
                (void) insQ(b.atRmv(i));
            else i++;
          break;
        case CL_DISCARD:
          for (i = 0; i < b.nodes(); )
            if (sense(s.select(*b[i]),reject))
              (void) b.atClr(i);
            else
              i++;
          break;
        case CL_CLONE:
          setFlags(CL_NEW|CL_DEL);
          for (i = 0; i < b.nodes(); i++)
            if (sense(s.select(*b[i]),reject))
              (void) insQNew(b[i]);
          break;
        default: assert(0); break;
    }
    return *this;
}

template <class KEY, class ITEM>
class CL_findAll : public CL<ITEM >
{
  public:
    CL_findAll<KEY,ITEM >& find(const CL<ITEM >& b,
        int (*cmpKey2Item)(const KEY * K,
        const ITEM *I), int honorSortFlag = 0,
        cl::RelCmp rc = EQ,
        CL_select_ops op = CL_SELECT);
    CL_find() : CL<ITEM >() {}
    CL_find(const CL<ITEM >& b, int (*cmpKey2Item)
        (const KEY * K, const ITEM * I),
        int honorSortFlag = 0,
        cl::RelCmp rc = EQ,
        CL_select_ops op = CL_SELECT)
        : CL<ITEM >()
        { (void) findAll(b,cmpKey2Item,
          honorSortFlag,rc,op); }
};

template <class KEY, class ITEM>
CL_findAll<KEY,ITEM >& CL_findAll<KEY,ITEM >::findAll
    (const CL<ITEM >& b, int (*cmpKey2Item)
    (const KEY * K, const ITEM * I),
    int honorSortFlag, cl::RelCmp rc, CL_select_ops op)
{
    unsigned idx;
    allClr();
    switch (op)  {
        case CL_SELECT:
          setFlags(CL_BIND_ONLY);
          if (b.findFirst(K,idx,CLKCC(ITEM)cmpKey2Item,
              honorSortFlag,rc))
              do {
                  (void) insQ(b[idx]);
              } while (b.findNext(K,idx,CLKCC(ITEM)
                  cmpKey2Item,honorSortFlag,rc));
          break;
        case CL_EXTRACT:
          setFlags(b.flags());
          if (b.findFirst(K,idx,CLKCC(ITEM)cmpKey2Item,
              honorSortFlag,rc))
              do {
                  if (flags(CL_DEL))
                      (void) insQDL(b.atRmv(idx--));
                  else
                      (void) insQ(b.atRmv(idx--));
              } while (b.findNext(K,idx,CLKCC(ITEM)
                  cmpKey2Item,honorSortFlag,rc));
          break;
        case CL_DISCARD:
          if (b.findFirst(K,idx,CLKCC(ITEM)cmpKey2Item,
              honorSortFlag,rc))
              do {
                  (void) b.atClr(idx--);
              } while (b.findNext(K,idx,CLKCC(ITEM)
                  cmpKey2Item,honorSortFlag,rc));
          break;
        case CL_CLONE:
          setFlags(CL_NEW|CL_DEL);
          if (b.findFirst(K,idx,CLKCC(ITEM)cmpKey2Item,
              honorSortFlag,rc))
              do {
                  (void) insQNew(b[idx]);
              } while (b.findNext(K,idx,CLKCC(ITEM)
                  cmpKey2Item,honorSortFlag,rc));
          break;
        default: assert(0); break;
    }
    return *this;
}

template <class ITEM, class XITEM>
class CL_collectBlock
{
    XITEM * (*__collect)(const ITEM& i);
  protected:
    friend CL_collect<ITEM >;
    virtual XITEM * collect(const ITEM& i)
        { assert(__collect); return (*__collect)(i); }
  public:
    CL_collectBlock
        (XITEM * (*_collect)(const ITEM& i) = 0)
        : __collect(_collect) {}
};

template <class ITEM, class XITEM>
class CL_collect : public CL<XITEM >
{
  public:
    CL_collect<ITEM,XITEM >& collect
        (const CL<ITEM >& b,
         CL_collectBlock<ITEM,XITEM >& c);
    CL_collect<ITEM,XITEM >& collect
        (const CL<ITEM >& b,
         XITEM * (*_collect)(const ITEM& i))
        { return collect(b,
          CL_collectBlock<ITEM,XITEM >(_collect)); }
    CL_collect() : CL<XITEM >() {}
    CL_collect(const CL<XITEM >& b,
        CL_collectBlock<ITEM >& c)
        : CL<ITEM >()
        { (void) collect(b,c); }
    CL_collect(const CL<ITEM >& b,
        XITEM * (*_collect)(const ITEM& i))
        : CL<ITEM >()
        { (void) collect(b,_collect); }
};

template <class ITEM, class XITEM>
CL_collect<ITEM,XITEM >&
CL_collect<ITEM,XITEM >::collect
    (const CL<ITEM >& b
    , CL_collectBlock<ITEM,XITEM >& c)
{
    allClr();
    setFlags(CL_NEW|CL_DEL);
    for (unsigned i = 0; i < b.nodes(); i++)
      (void) insQDL(c.collect(*b[i]));
    return *this;
}

#endif  /*  !CL_NO_TEMPLATES  */
#endif  /*  cl_h  */
