/*
        single.h -- allow only a single instance
          per <SCOPE,TYPE,ID> tuple!
        (C) Copyright 1995  John Webster Small
        All rights reserved

        TYPE must have a default constructor.
        Singletons of arrays are not allowed!

          Singleton<GLOBAL_SCOPE,int> onlyInt;
          (int &) onlyInt = 5;
*/

#ifndef single_h
#define single_h

/* LINTLIBRARY */

#ifndef assert_h
  #include <assert.h>
  #ifndef assert_h
    #define assert_h
  #endif
#endif

#ifndef limits_h
  #include <limits.h>
  #ifndef limits_h
    #define limits_h
  #endif
#endif

#ifndef GLOBAL_SCOPE
  #define SCOPE_TYPE(scope) scope ## _SCOPE
  #define SCOPE_TYPEDEF(scope) \
    struct SCOPE_TYPE(scope) {}
  SCOPE_TYPEDEF(global);
  #define GLOBAL_SCOPE SCOPE_TYPE(global)
#endif

template <class SCOPE, class TYPE, int ID = 0>
struct Singleton_traits
{
    static TYPE * newInstance() { return new TYPE; }
};

enum SingletonOp
    { SGLTN_RELEASE, SGLTN_CREATE, SGLTN_PRESENT };

template <class SCOPE, class TYPE, int ID = 0>
class Singleton_Base
{
  public:
    unsigned long refs(int inc = 0) const;
    TYPE * instance
        (SingletonOp op = SGLTN_CREATE) const;
    Singleton_Base() { (void) refs(1); }
    ~Singleton_Base()
        { if (!refs(-1))
          (void) instance(SGLTN_RELEASE); }
};

template <class SCOPE, class TYPE, int ID>
unsigned long
Singleton_Base<SCOPE,TYPE,ID>::refs(int inc) const
{
    static unsigned long _refs = 0;
    if (inc > 0)  {
        assert(_refs < ULONG_MAX);
        _refs++;
    }
    else if (inc < 0)  {
        assert(_refs);
        _refs--;
    }
    return _refs;
}

template <class SCOPE, class TYPE, int ID>
TYPE *
Singleton_Base<SCOPE,TYPE,ID>::
    instance(SingletonOp op) const
{
    static TYPE * _instance = 0;
    if (op != SGLTN_RELEASE && !refs()) return 0;
    if (op == SGLTN_RELEASE && _instance)  {
        delete _instance;
        _instance = 0;
    }
    if (op == SGLTN_CREATE && !_instance)
        _instance = Singleton_traits<SCOPE,TYPE,ID>::
            newInstance();
    return _instance;
};

template <class SCOPE, class TYPE, int ID = 0>
class Singleton : Singleton_Base<SCOPE,TYPE,ID>
{
  public:
    Singleton() : Singleton_Base<SCOPE,TYPE,ID>() {}
    unsigned long refs() const { return
        Singleton_Base<SCOPE,TYPE,ID>::refs(); }
    operator TYPE * ()
        { return instance(SGLTN_CREATE); }
    TYPE * ok() const
        { return instance(SGLTN_PRESENT); }
    operator TYPE & ()
        { TYPE * T = instance(SGLTN_CREATE);
          assert(T);  return *T; }
    void release() { (void) instance(SGLTN_RELEASE); }
};

#endif
